/*
   Simple device driver for access to the i2c bus (pending getting full linux driver working)
   Intended to access the AIDA eeprom (Microchip 24AA02E48) and extract the ethernet MAC address
*/


#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>   /* printk() */
#include <linux/fs.h>       /* everything... */
#include <linux/ioport.h>
#include <linux/slab.h>

// #include <asm/processor.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include "iic.h"

MODULE_AUTHOR("STFC Daresbury Laboratory");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("IIC Device Driver");

//MODULE_PARM(debug_level, "i");

#define DEVICE_NAME "iic"

static struct file_operations iic_ops = {
    .owner   = THIS_MODULE,
    .open    = iic_open,
    .release = iic_release,
    .read    = iic_read,
    .write   = iic_write,
    .ioctl   = iic_ioctl
};

static unsigned int iic_major = 0;  /* use dynamic allocation */

unsigned long base0 = 0x81600000;
unsigned long size0 = 0x00000200;    /*  512 bytes  */
unsigned long *base0_mapped;


/*
 * Open the device
*/

static int iic_open (struct inode *inode, struct file *filp)
{

/*    claim all resources  */

    DPRINTF("iic_open:\n");
        
    return 0;
}

/*
 * Close the device
*/

static int iic_release (struct inode *inode, struct file *filp)
{

/*    free all resources  */

    DPRINTF("iic_release:\n");
       
    return 0;
}

/*
 *  device ioctl
*/

static int iic_ioctl (struct inode *inode, struct file *filp, unsigned int cmd,
     unsigned long arg)
{

    DPRINTF("iic_ioctl: called with command %d\n", cmd);
    
    switch (cmd)  {
    
       default:
         return -ENOTTY;
    }
    
    return(0);
}


/*
 *  device read
*/

static ssize_t iic_read (struct file *filp,
     char *buffer,     /* the buffer to fill with data  */
     size_t length,    /* the length of the buffer      */
     loff_t *offset)   /* offset in the file            */
{
    int count;
    unsigned long *srcptr;          /* pointer in data buffer  */
    unsigned long *destptr;         /* address of hardware */
    int rc = length;

    unsigned char * kbuf = kmalloc(length, GFP_KERNEL);
    srcptr = (unsigned long *)kbuf;
    
    count = length >> 2;    /* number of 32 bit words to read */

    destptr = (unsigned long*) ((unsigned char*)base0_mapped + *offset);
    while (count--) {*srcptr++ = *destptr++;}

    if (copy_to_user(buffer, kbuf, length)) rc = -EFAULT;            

    kfree(kbuf);

    return rc;
}

/*
 *  device write
*/

static ssize_t iic_write (struct file *filp,
     const char *buffer,     /* the data buffer */
     size_t length,    /* the length of the buffer      */
     loff_t *offset)   /* offset in the file            */
{
    int count;
    unsigned long *srcptr;          /* pointer in data buffer  */
    unsigned long *destptr;         /* address of hardware */
    int rc = length;
    
    unsigned char * kbuf = kmalloc(length, GFP_KERNEL);
    
    if (copy_from_user(kbuf, buffer, length)) return -EFAULT;
    srcptr = (unsigned long *)kbuf;
        
    count = length >> 2;    /* number of 32 bit words to write */

    destptr = (unsigned long*) ((unsigned char*)base0_mapped + *offset);
    while (count--) {*destptr++ = *srcptr++;}
    
    kfree(kbuf);
            
    return rc;
}


/*
 * Module housekeeping.
*/
 
/*
     iic_init  -  module initialisation
     entered by the kernel when the module is loaded
*/
 
static int __init iic_init(void)
{
    struct resource* res;

//    obtain region of address space required

    res = request_mem_region(base0, size0, DEVICE_NAME);  /* the return pointer is meaningless */
    if (res == NULL) {
            printk(KERN_INFO "iic: can't get I/O mem address 0x%lx\n",base0);
            return -ENODEV;
    }

//     and ioremap it

    base0_mapped = (unsigned long*) ioremap_nocache(base0, size0);
    
    DPRINTF("iic: mem region start 0x%lx for 0x%lx mapped at 0x%lx\n", base0, size0,(unsigned long) base0_mapped);
            
//    register the device

    iic_major = register_chrdev(iic_major, DEVICE_NAME, &iic_ops);
    if (iic_major < 0) {
        printk(KERN_WARNING "iic: registering the device failed with %d\n", iic_major);
        return iic_major;
    }
    
    DPRINTF("iic: driver assigned major number %d\n", iic_major);
    
    return 0;
}

/*
    iic_cleanup  -  module cleanup
    entered by the kernel when the module is unloaded
*/

static void __exit iic_cleanup(void)
{

    DPRINTF("iic module is exiting\n");
    
    unregister_chrdev(iic_major, DEVICE_NAME);
    
    iounmap((void *) base0_mapped);
    release_mem_region(base0,size0);
    
}


//    define module initialisation and termination procedures

module_init(iic_init);
module_exit(iic_cleanup);



